@@ -14,6 +14,9 @@
     #define PYUV_PYTHON3
     #define PyInt_FromSsize_t PyLong_FromSsize_t
     #define PyInt_FromLong PyLong_FromLong
+    #define PYUV_BYTES "y"
+#else
+    #define PYUV_BYTES "s"
 #endif
 
 /* libuv */
@@ -27,15 +30,17 @@ typedef int Bool;
 
 
 /* Utility macros */
-#ifndef __STRING
-    #define __STRING(x) #x
-#endif
-#define __MSTR(x) __STRING(x)
+#define PYUV_STRINGIFY_HELPER(x) #x
+#define PYUV_STRINGIFY(x) PYUV_STRINGIFY_HELPER(x)
+
+#define PYUV_CONTAINER_OF(ptr, type, field)                                  \
+      ((type *) ((char *) (ptr) - ((long) &((type *) 0)->field)))
 
 #define UNUSED_ARG(arg)  (void)arg
 
 #if defined(__MINGW32__) || defined(_MSC_VER)
     #define PYUV_WINDOWS
+    #define PYUV_MAXSTDIO 2048
 #endif
 
 #ifdef _MSC_VER
@@ -44,11 +49,6 @@ typedef int Bool;
     #define INLINE inline
 #endif
 
-/* borrowed from pyev */
-#ifdef PYUV_WINDOWS
-    #define PYUV_MAXSTDIO 2048
-#endif
-
 #define ASSERT(x)                                                           \
     do {                                                                    \
         if (!(x)) {                                                         \
@@ -66,6 +66,10 @@ typedef int Bool;
 
 #define UV_HANDLE_LOOP(x) UV_LOOP(HANDLE(x))
 
+#define REQUEST(x) ((Request *)x)
+
+#define UV_REQUEST(x) REQUEST(x)->req_ptr
+
 #define RAISE_IF_INITIALIZED(obj, retval)                                           \
     do {                                                                            \
         if ((obj)->initialized) {                                                   \
@@ -124,12 +128,12 @@ static PyTypeObject LoopType;
 /* Handle */
 typedef struct {
     PyObject_HEAD
+    uv_handle_t *uv_handle;
     Bool initialized;
     PyObject *weakreflist;
     PyObject *dict;
     Loop *loop;
     PyObject *on_close_cb;
-    uv_handle_t *uv_handle;
 } Handle;
 
 static PyTypeObject HandleType;
@@ -137,6 +141,7 @@ static PyTypeObject HandleType;
 /* Async */
 typedef struct {
     Handle handle;
+    uv_async_t async_h;
     PyObject *callback;
 } Async;
 
@@ -145,6 +150,7 @@ static PyTypeObject AsyncType;
 /* Timer */
 typedef struct {
     Handle handle;
+    uv_timer_t timer_h;
     PyObject *callback;
 } Timer;
 
@@ -153,6 +159,7 @@ static PyTypeObject TimerType;
 /* Prepare */
 typedef struct {
     Handle handle;
+    uv_prepare_t prepare_h;
     PyObject *callback;
 } Prepare;
 
@@ -161,6 +168,7 @@ static PyTypeObject PrepareType;
 /* Idle */
 typedef struct {
     Handle handle;
+    uv_idle_t idle_h;
     PyObject *callback;
 } Idle;
 
@@ -169,6 +177,7 @@ static PyTypeObject IdleType;
 /* Check */
 typedef struct {
     Handle handle;
+    uv_check_t check_h;
     PyObject *callback;
 } Check;
 
@@ -177,6 +186,7 @@ static PyTypeObject CheckType;
 /* Signal */
 typedef struct {
     Handle handle;
+    uv_signal_t signal_h;
     PyObject *callback;
 } Signal;
 
@@ -185,6 +195,7 @@ static PyTypeObject SignalType;
 /* SignalChecker */
 typedef struct {
     Handle handle;
+    uv_poll_t poll_h;
     long fd;
 } SignalChecker;
 
@@ -201,6 +212,7 @@ static PyTypeObject StreamType;
 /* TCP */
 typedef struct {
     Stream stream;
+    uv_tcp_t tcp_h;
     PyObject *on_new_connection_cb;
 } TCP;
 
@@ -209,6 +221,7 @@ static PyTypeObject TCPType;
 /* Pipe */
 typedef struct {
     Stream stream;
+    uv_pipe_t pipe_h;
     PyObject *on_new_connection_cb;
 } Pipe;
 
@@ -217,6 +230,7 @@ static PyTypeObject PipeType;
 /* TTY */
 typedef struct {
     Stream stream;
+    uv_tty_t tty_h;
 } TTY;
 
 static PyTypeObject TTYType;
@@ -224,6 +238,7 @@ static PyTypeObject TTYType;
 /* UDP */
 typedef struct {
     Handle handle;
+    uv_udp_t udp_h;
     PyObject *on_read_cb;
 } UDP;
 
@@ -232,6 +247,7 @@ static PyTypeObject UDPType;
 /* Poll */
 typedef struct {
     Handle handle;
+    uv_poll_t poll_h;
     PyObject *callback;
     long fd;
 } Poll;
@@ -251,6 +267,7 @@ static PyTypeObject StdIOType;
 typedef struct {
     Handle handle;
     Bool spawned;
+    uv_process_t process_h;
     PyObject *on_exit_cb;
     PyObject *stdio;
 } Process;
@@ -260,6 +277,7 @@ static PyTypeObject ProcessType;
 /* FSEvent */
 typedef struct {
     Handle handle;
+    uv_fs_event_t fsevent_h;
     PyObject *callback;
 } FSEvent;
 
@@ -268,6 +286,7 @@ static PyTypeObject FSEventType;
 /* FSPoll */
 typedef struct {
     Handle handle;
+    uv_fs_poll_t fspoll_h;
     PyObject *callback;
 } FSPoll;
 
@@ -321,7 +340,8 @@ static PyTypeObject SemaphoreType;
 /* Request */
 typedef struct {
     PyObject_HEAD
-    uv_req_t *req;
+    uv_req_t *req_ptr;
+    Loop *loop;
 } Request;
 
 static PyTypeObject RequestType;
@@ -329,6 +349,7 @@ static PyTypeObject RequestType;
 /* GAIRequest */
 typedef struct {
     Request request;
+    uv_getaddrinfo_t req;
     PyObject *callback;
 } GAIRequest;
 
@@ -337,6 +358,7 @@ static PyTypeObject GAIRequestType;
 /* WorkRequest */
 typedef struct {
     Request request;
+    uv_work_t req;
     PyObject *work_cb;
     PyObject *done_cb;
 } WorkRequest;
@@ -346,6 +368,7 @@ static PyTypeObject WorkRequestType;
 /* FSRequest */
 typedef struct {
     Request request;
+    uv_fs_t req;
     PyObject *callback;
 } FSRequest;
 
@@ -552,13 +575,7 @@ pyseq2uvbuf(PyObject *seq, Py_buffer **rviews, uv_buf_t **rbufs, int *rbuf_count
     }
 
     for (i = 0; i < buf_count; i++) {
-        if (!PyArg_Parse(PySequence_Fast_GET_ITEM(data_fast, i),
-#ifdef PYUV_PYTHON3
-                         "y*;argument 1 must be an iterable of buffer-compatible objects",
-#else
-                         "s*;argument 1 must be an iterable of buffer-compatible objects",
-#endif
-                         &views[i])) {
+        if (!PyArg_Parse(PySequence_Fast_GET_ITEM(data_fast, i), PYUV_BYTES"*;argument 1 must be an iterable of buffer-compatible objects", &views[i])) {
             goto error;
         }
         uv_bufs[i].base = views[i].buf;
@@ -582,26 +599,95 @@ error:
 }
 
 
-/* guess IP address family */
-static INLINE int
-pyuv_guess_ip_family(char *ip, int *address_type)
+/* parse a Python tuple containing host, port, flowinfo, scope_id into a sockaddr struct */
+static int
+pyuv_parse_addr_tuple(PyObject *addr, struct sockaddr *sa)
 {
+    char *host;
+    int port;
+    unsigned int scope_id, flowinfo;
     struct in_addr addr4;
     struct in6_addr addr6;
-    uv_err_t err;
+    struct sockaddr_in *sa4;
+    struct sockaddr_in6 *sa6;
 
-    err = uv_inet_pton(AF_INET, ip, &addr4);
-    if (err.code == UV_OK) {
-        *address_type = AF_INET;
+    flowinfo = scope_id = 0;
+
+    if (!PyTuple_Check(addr)) {
+        PyErr_Format(PyExc_TypeError, "address must be tuple, not %.500s", Py_TYPE(addr)->tp_name);
+        return -1;
+    }
+
+    if (!PyArg_ParseTuple(addr, "si|II", &host, &port, &flowinfo, &scope_id)) {
+        return -1;
+    }
+
+    if (port < 0 || port > 0xffff) {
+        PyErr_SetString(PyExc_OverflowError, "port must be 0-65535");
+        return -1;
+    }
+
+    if (flowinfo > 0xfffff) {
+        PyErr_SetString(PyExc_OverflowError, "flowinfo must be 0-1048575");
+        return -1;
+    }
+
+    memset(sa, 0, sizeof(struct sockaddr));
+
+    if (uv_inet_pton(AF_INET, host, &addr4).code == UV_OK) {
+        /* it's an IPv4 address */
+        sa4 = (struct sockaddr_in *)sa;
+        sa4->sin_family = AF_INET;
+        sa4->sin_port = htons((short)port);
+        sa4->sin_addr = addr4;
+        return 0;
+    } else if (uv_inet_pton(AF_INET6, host, &addr6).code == UV_OK) {
+        /* it's an IPv4 address */
+        sa6 = (struct sockaddr_in6 *)sa;
+        sa6->sin6_family = AF_INET6;
+        sa6->sin6_port = htons((short)port);
+        sa6->sin6_addr = addr6;
+        sa6->sin6_flowinfo = flowinfo;
+        sa6->sin6_scope_id = scope_id;
         return 0;
     } else {
-        err = uv_inet_pton(AF_INET6, ip, &addr6);
-        if (err.code == UV_OK) {
-            *address_type = AF_INET6;
-            return 0;
-        } else {
-            return 1;
-        }
+        PyErr_SetString(PyExc_ValueError, "invalid IP address");
+        return -1;
+    }
+}
+
+
+/* Modified from Python Modules/socketmodule.c */
+static PyObject *
+makesockaddr(struct sockaddr *addr, int addrlen)
+{
+    static char buf[INET6_ADDRSTRLEN+1];
+    struct sockaddr_in *addr4;
+    struct sockaddr_in6 *addr6;
+
+    if (addrlen == 0) {
+        /* No address */
+        Py_RETURN_NONE;
+    }
+
+    switch (addr->sa_family) {
+    case AF_INET:
+    {
+        addr4 = (struct sockaddr_in*)addr;
+        uv_ip4_name(addr4, buf, sizeof(buf));
+        return Py_BuildValue("si", buf, ntohs(addr4->sin_port));
+    }
+
+    case AF_INET6:
+    {
+        addr6 = (struct sockaddr_in6*)addr;
+        uv_ip6_name(addr6, buf, sizeof(buf));
+        return Py_BuildValue("siII", buf, ntohs(addr6->sin6_port), ntohl(addr6->sin6_flowinfo), addr6->sin6_scope_id);
+    }
+
+    default:
+        /* If we don't know the address family, don't raise an exception -- return it as a tuple. */
+        return Py_BuildValue("is#", addr->sa_family, addr->sa_data, sizeof(addr->sa_data));
     }
 }
 
