@@ -1,12 +1,21 @@
 
 typedef struct {
+    uv_write_t req;
+    Stream *obj;
     PyObject *callback;
     PyObject *send_handle;
     Py_buffer *views;
+    Py_buffer view[1];
     int view_count;
-} stream_write_data_t;
+} stream_write_ctx;
 
 
+typedef struct {
+    uv_shutdown_t req;
+    Stream *obj;
+    PyObject *callback;
+} stream_shutdown_ctx;
+
 static uv_buf_t
 on_stream_alloc(uv_stream_t* handle, size_t suggested_size)
 {
@@ -20,18 +29,17 @@ static void
 on_stream_shutdown(uv_shutdown_t* req, int status)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
-    uv_err_t err;
+    stream_shutdown_ctx *ctx;
     Stream *self;
     PyObject *callback, *result, *py_errorno;
 
-    self = (Stream *)req->handle->data;
-    callback = (PyObject *)req->data;
-
-    ASSERT(self);
+    ctx = PYUV_CONTAINER_OF(req, stream_shutdown_ctx, req);
+    self = ctx->obj;
+    callback = ctx->callback;
 
     if (callback != Py_None) {
         if (status < 0) {
-            err = uv_last_error(UV_HANDLE_LOOP(self));
+            uv_err_t err = uv_last_error(UV_HANDLE_LOOP(self));
             py_errorno = PyInt_FromLong((long)err.code);
         } else {
             py_errorno = Py_None;
@@ -64,8 +72,9 @@ on_stream_read(uv_stream_t* handle, int nread, uv_buf_t buf)
     PyObject *result, *data, *py_errorno;
     ASSERT(handle);
 
+    /* Can't use container_of here */
     self = (Stream *)handle->data;
-    ASSERT(self);
+
     /* Object could go out of scope in the callback, increase refcount to avoid it */
     Py_INCREF(self);
 
@@ -100,19 +109,17 @@ on_stream_write(uv_write_t* req, int status)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
     int i;
-    stream_write_data_t* req_data;
+    stream_write_ctx *ctx;
     Stream *self;
     PyObject *callback, *send_handle, *result, *py_errorno;
     uv_err_t err;
 
     ASSERT(req);
 
-    req_data = (stream_write_data_t *)req->data;
-    self = (Stream *)req->handle->data;
-    callback = req_data->callback;
-    send_handle = req_data->send_handle;
-
-    ASSERT(self);
+    ctx = PYUV_CONTAINER_OF(req, stream_write_ctx, req);
+    self = ctx->obj;
+    callback = ctx->callback;
+    send_handle = ctx->send_handle;
 
     if (callback != Py_None) {
         if (status < 0) {
@@ -132,12 +139,13 @@ on_stream_write(uv_write_t* req, int status)
 
     Py_DECREF(callback);
     Py_XDECREF(send_handle);
-    for (i = 0; i < req_data->view_count; i++) {
-        PyBuffer_Release(&req_data->views[i]);
+    for (i = 0; i < ctx->view_count; i++) {
+        PyBuffer_Release(&ctx->views[i]);
     }
-    PyMem_Free(req_data->views);
-    PyMem_Free(req_data);
-    PyMem_Free(req);
+    if (ctx->views != ctx->view) {
+        PyMem_Free(ctx->views);
+    }
+    PyMem_Free(ctx);
 
     /* Refcount was increased in the caller function */
     Py_DECREF(self);
@@ -150,8 +158,11 @@ static PyObject *
 Stream_func_shutdown(Stream *self, PyObject *args)
 {
     int r;
-    uv_shutdown_t *req = NULL;
-    PyObject *callback = Py_None;
+    stream_shutdown_ctx *ctx;
+    PyObject *callback;
+
+    ctx = NULL;
+    callback = Py_None;
 
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
@@ -160,17 +171,18 @@ Stream_func_shutdown(Stream *self, PyObject *args)
         return NULL;
     }
 
-    Py_INCREF(callback);
-
-    req = PyMem_Malloc(sizeof *req);
-    if (!req) {
+    ctx = PyMem_Malloc(sizeof *ctx);
+    if (!ctx) {
         PyErr_NoMemory();
-        goto error;
+        return NULL;
     }
 
-    req->data = (void *)callback;
+    Py_INCREF(callback);
 
-    r = uv_shutdown(req, (uv_stream_t *)UV_HANDLE(self), on_stream_shutdown);
+    ctx->obj = self;
+    ctx->callback = callback;
+
+    r = uv_shutdown(&ctx->req, (uv_stream_t *)UV_HANDLE(self), on_stream_shutdown);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_StreamError);
         goto error;
@@ -183,7 +195,7 @@ Stream_func_shutdown(Stream *self, PyObject *args)
 
 error:
     Py_DECREF(callback);
-    PyMem_Free(req);
+    PyMem_Free(ctx);
     return NULL;
 }
 
@@ -245,38 +257,23 @@ Stream_func_stop_read(Stream *self)
 
 
 static INLINE PyObject *
-pyuv_stream_write(Stream *self, Py_buffer *views, uv_buf_t *bufs, int buf_count, PyObject *callback, PyObject *send_handle)
+pyuv_stream_write(Stream *self, stream_write_ctx *ctx, Py_buffer *views, uv_buf_t *bufs, int buf_count, PyObject *callback, PyObject *send_handle)
 {
     int i, r;
-    uv_write_t *wr = NULL;
-    stream_write_data_t *req_data = NULL;
 
     Py_INCREF(callback);
     Py_XINCREF(send_handle);
 
-    wr = PyMem_Malloc(sizeof *wr);
-    if (!wr) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    req_data = PyMem_Malloc(sizeof *req_data);
-    if (!req_data) {
-        PyErr_NoMemory();
-        goto error;
-    }
-
-    req_data->callback = callback;
-    req_data->send_handle = send_handle;
-    req_data->views = views;
-    req_data->view_count = buf_count;
-
-    wr->data = (void *)req_data;
+    ctx->obj = self;
+    ctx->callback = callback;
+    ctx->send_handle = send_handle;
+    ctx->views = views;
+    ctx->view_count = buf_count;
 
     if (send_handle) {
-        r = uv_write2(wr, (uv_stream_t *)UV_HANDLE(self), bufs, buf_count, (uv_stream_t *)UV_HANDLE(send_handle), on_stream_write);
+        r = uv_write2(&ctx->req, (uv_stream_t *)UV_HANDLE(self), bufs, buf_count, (uv_stream_t *)UV_HANDLE(send_handle), on_stream_write);
     } else {
-        r = uv_write(wr, (uv_stream_t *)UV_HANDLE(self), bufs, buf_count, on_stream_write);
+        r = uv_write(&ctx->req, (uv_stream_t *)UV_HANDLE(self), bufs, buf_count, on_stream_write);
     }
 
     if (r != 0) {
@@ -295,9 +292,10 @@ error:
     for (i = 0; i < buf_count; i++) {
         PyBuffer_Release(&views[i]);
     }
-    PyMem_Free(views);
-    PyMem_Free(req_data);
-    PyMem_Free(wr);
+    if (ctx->views != ctx->view) {
+        PyMem_Free(views);
+    }
+    PyMem_Free(ctx);
     return NULL;
 }
 
@@ -306,36 +304,36 @@ static PyObject *
 Stream_func_write(Stream *self, PyObject *args)
 {
     uv_buf_t buf;
+    stream_write_ctx *ctx;
     Py_buffer *view;
     PyObject *callback = Py_None;
 
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    view = PyMem_Malloc(sizeof *view);
-    if (!view) {
+    ctx = PyMem_Malloc(sizeof *ctx);
+    if (!ctx) {
         PyErr_NoMemory();
         return NULL;
     }
 
-#ifdef PYUV_PYTHON3
-    if (!PyArg_ParseTuple(args, "y*|O:write", view, &callback)) {
-#else
-    if (!PyArg_ParseTuple(args, "s*|O:write", view, &callback)) {
-#endif
+    view = &ctx->view[0];
+
+    if (!PyArg_ParseTuple(args, PYUV_BYTES"*|O:write", view, &callback)) {
+        PyMem_Free(ctx);
         return NULL;
     }
 
     if (callback != Py_None && !PyCallable_Check(callback)) {
         PyBuffer_Release(view);
-        PyMem_Free(view);
+        PyMem_Free(ctx);
         PyErr_SetString(PyExc_TypeError, "a callable or None is required");
         return NULL;
     }
 
     buf = uv_buf_init(view->buf, view->len);
 
-    return pyuv_stream_write(self, view, &buf, 1, callback, NULL);
+    return pyuv_stream_write(self, ctx, view, &buf, 1, callback, NULL);
 }
 
 
@@ -346,6 +344,7 @@ Stream_func_writelines(Stream *self, PyObject *args)
     Py_buffer *views;
     PyObject *callback, *seq, *ret;
     uv_buf_t *bufs;
+    stream_write_ctx *ctx;
 
     callback = Py_None;
 
@@ -361,13 +360,20 @@ Stream_func_writelines(Stream *self, PyObject *args)
         return NULL;
     }
 
-    r = pyseq2uvbuf(seq, &views, &bufs, &buf_count);
-    if (r != 0) {
-        /* error is already set */
+    ctx = PyMem_Malloc(sizeof *ctx);
+    if (!ctx) {
+        PyErr_NoMemory();
         return NULL;
     }
 
-    ret = pyuv_stream_write(self, views, bufs, buf_count, callback, NULL);
+    r = pyseq2uvbuf(seq, &views, &bufs, &buf_count);
+    if (r != 0) {
+        /* error is already set */
+        PyMem_Free(ctx);
+        return NULL;
+    }
+
+    ret = pyuv_stream_write(self, ctx, views, bufs, buf_count, callback, NULL);
 
     /* uv_write copies the uv_buf_t structures, so we can free them now */
     PyMem_Free(bufs);
