@@ -1,15 +1,15 @@
 
 static void
-on_tcp_connection(uv_stream_t* server, int status)
+on_tcp_connection(uv_stream_t *handle, int status)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
     TCP *self;
     PyObject *result, *py_errorno;
 
-    ASSERT(server);
-    self = (TCP *)server->data;
+    ASSERT(handle);
+
+    self = PYUV_CONTAINER_OF(handle, TCP, tcp_h);
 
-    ASSERT(self);
     /* Object could go out of scope in the callback, increase refcount to avoid it */
     Py_INCREF(self);
 
@@ -41,11 +41,9 @@ on_tcp_client_connection(uv_connect_t *req, int status)
     PyObject *callback, *result, *py_errorno;
 
     ASSERT(req);
-    self = (TCP *)req->handle->data;
+    self = PYUV_CONTAINER_OF(req->handle, TCP, tcp_h);
     callback = (PyObject *)req->data;
 
-    ASSERT(self);
-
     if (status != 0) {
         uv_err_t err = uv_last_error(UV_HANDLE_LOOP(self));
         py_errorno = PyInt_FromLong(err.code);
@@ -74,30 +72,26 @@ on_tcp_client_connection(uv_connect_t *req, int status)
 static PyObject *
 TCP_func_bind(TCP *self, PyObject *args)
 {
-    int r, bind_port, address_type;
-    char *bind_ip;
+    int r;
+    struct sockaddr sa;
+    PyObject *addr;
 
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    if (!PyArg_ParseTuple(args, "(si):bind", &bind_ip, &bind_port)) {
+    if (!PyArg_ParseTuple(args, "O:bind", &addr)) {
         return NULL;
     }
 
-    if (bind_port < 0 || bind_port > 65535) {
-        PyErr_SetString(PyExc_ValueError, "port must be between 0 and 65535");
+    if (pyuv_parse_addr_tuple(addr, &sa) < 0) {
+        /* Error is set by the function itself */
         return NULL;
     }
 
-    if (pyuv_guess_ip_family(bind_ip, &address_type)) {
-        PyErr_SetString(PyExc_ValueError, "invalid IP address");
-        return NULL;
-    }
-
-    if (address_type == AF_INET) {
-        r = uv_tcp_bind((uv_tcp_t *)UV_HANDLE(self), uv_ip4_addr(bind_ip, bind_port));
+    if (sa.sa_family == AF_INET) {
+        r = uv_tcp_bind(&self->tcp_h, *(struct sockaddr_in *)&sa);
     } else {
-        r = uv_tcp_bind6((uv_tcp_t *)UV_HANDLE(self), uv_ip6_addr(bind_ip, bind_port));
+        r = uv_tcp_bind6(&self->tcp_h, *(struct sockaddr_in6 *)&sa);
     }
 
     if (r != 0) {
@@ -135,7 +129,7 @@ TCP_func_listen(TCP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_listen((uv_stream_t *)UV_HANDLE(self), backlog, on_tcp_connection);
+    r = uv_listen((uv_stream_t *)&self->tcp_h, backlog, on_tcp_connection);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_TCPError);
         return NULL;
@@ -168,7 +162,7 @@ TCP_func_accept(TCP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_accept((uv_stream_t *)UV_HANDLE(self), (uv_stream_t *)UV_HANDLE(client));
+    r = uv_accept((uv_stream_t *)&self->tcp_h, (uv_stream_t *)UV_HANDLE(client));
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_TCPError);
         return NULL;
@@ -181,15 +175,15 @@ TCP_func_accept(TCP *self, PyObject *args)
 static PyObject *
 TCP_func_connect(TCP *self, PyObject *args)
 {
-    int r, connect_port, address_type;
-    char *connect_ip;
+    int r;
+    struct sockaddr sa;
     uv_connect_t *connect_req = NULL;
-    PyObject *callback;
+    PyObject *addr, *callback;
 
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    if (!PyArg_ParseTuple(args, "(si)O:connect", &connect_ip, &connect_port, &callback)) {
+    if (!PyArg_ParseTuple(args, "OO:connect", &addr, &callback)) {
         return NULL;
     }
 
@@ -198,13 +192,8 @@ TCP_func_connect(TCP *self, PyObject *args)
         return NULL;
     }
 
-    if (connect_port < 0 || connect_port > 65535) {
-        PyErr_SetString(PyExc_ValueError, "port must be between 0 and 65535");
-        return NULL;
-    }
-
-    if (pyuv_guess_ip_family(connect_ip, &address_type)) {
-        PyErr_SetString(PyExc_ValueError, "invalid IP address");
+    if (pyuv_parse_addr_tuple(addr, &sa) < 0) {
+        /* Error is set by the function itself */
         return NULL;
     }
 
@@ -216,12 +205,12 @@ TCP_func_connect(TCP *self, PyObject *args)
         goto error;
     }
 
-    connect_req->data = (void *)callback;
+    connect_req->data = callback;
 
-    if (address_type == AF_INET) {
-        r = uv_tcp_connect(connect_req, (uv_tcp_t *)UV_HANDLE(self), uv_ip4_addr(connect_ip, connect_port), on_tcp_client_connection);
+    if (sa.sa_family == AF_INET) {
+        r = uv_tcp_connect(connect_req, &self->tcp_h, *(struct sockaddr_in *)&sa, on_tcp_client_connection);
     } else {
-        r = uv_tcp_connect6(connect_req, (uv_tcp_t *)UV_HANDLE(self), uv_ip6_addr(connect_ip, connect_port), on_tcp_client_connection);
+        r = uv_tcp_connect6(connect_req, &self->tcp_h, *(struct sockaddr_in6 *)&sa, on_tcp_client_connection);
     }
 
     if (r != 0) {
@@ -245,34 +234,20 @@ static PyObject *
 TCP_func_getsockname(TCP *self)
 {
     int r, namelen;
-    char ip[INET6_ADDRSTRLEN];
     struct sockaddr sockname;
-    struct sockaddr_in *addr4;
-    struct sockaddr_in6 *addr6;
 
     namelen = sizeof(sockname);
 
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    r = uv_tcp_getsockname((uv_tcp_t *)UV_HANDLE(self), &sockname, &namelen);
+    r = uv_tcp_getsockname(&self->tcp_h, &sockname, &namelen);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_TCPError);
         return NULL;
     }
 
-    if (sockname.sa_family == AF_INET) {
-        addr4 = (struct sockaddr_in*)&sockname;
-        uv_ip4_name(addr4, ip, INET_ADDRSTRLEN);
-        return Py_BuildValue("si", ip, ntohs(addr4->sin_port));
-    } else if (sockname.sa_family == AF_INET6) {
-        addr6 = (struct sockaddr_in6*)&sockname;
-        uv_ip6_name(addr6, ip, INET6_ADDRSTRLEN);
-        return Py_BuildValue("si", ip, ntohs(addr6->sin6_port));
-    } else {
-        PyErr_SetString(PyExc_TCPError, "unknown address type detected");
-        return NULL;
-    }
+    return makesockaddr(&sockname, namelen);
 }
 
 
@@ -280,36 +255,20 @@ static PyObject *
 TCP_func_getpeername(TCP *self)
 {
     int r, namelen;
-    char ip[INET6_ADDRSTRLEN];
     struct sockaddr peername;
-    struct sockaddr_in *addr4;
-    struct sockaddr_in6 *addr6;
 
     namelen = sizeof(peername);
 
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    r = uv_tcp_getpeername((uv_tcp_t *)UV_HANDLE(self), &peername, &namelen);
+    r = uv_tcp_getpeername(&self->tcp_h, &peername, &namelen);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_TCPError);
         return NULL;
     }
 
-    if (peername.sa_family == AF_INET) {
-        addr4 = (struct sockaddr_in*)&peername;
-        r = uv_ip4_name(addr4, ip, INET_ADDRSTRLEN);
-        ASSERT(r == 0);
-        return Py_BuildValue("si", ip, ntohs(addr4->sin_port));
-    } else if (peername.sa_family == AF_INET6) {
-        addr6 = (struct sockaddr_in6*)&peername;
-        r = uv_ip6_name(addr6, ip, INET6_ADDRSTRLEN);
-        ASSERT(r == 0);
-        return Py_BuildValue("si", ip, ntohs(addr6->sin6_port));
-    } else {
-        PyErr_SetString(PyExc_TCPError, "unknown address type detected");
-        return NULL;
-    }
+    return makesockaddr(&peername, namelen);
 }
 
 
@@ -326,7 +285,7 @@ TCP_func_nodelay(TCP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_tcp_nodelay((uv_tcp_t *)UV_HANDLE(self), (enable == Py_True) ? 1 : 0);
+    r = uv_tcp_nodelay(&self->tcp_h, (enable == Py_True) ? 1 : 0);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_TCPError);
         return NULL;
@@ -350,7 +309,7 @@ TCP_func_keepalive(TCP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_tcp_keepalive((uv_tcp_t *)UV_HANDLE(self), (enable == Py_True) ? 1 : 0, delay);
+    r = uv_tcp_keepalive(&self->tcp_h, (enable == Py_True) ? 1 : 0, delay);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_TCPError);
         return NULL;
@@ -373,7 +332,7 @@ TCP_func_simultaneous_accepts(TCP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_tcp_simultaneous_accepts((uv_tcp_t *)UV_HANDLE(self), (enable == Py_True) ? 1 : 0);
+    r = uv_tcp_simultaneous_accepts(&self->tcp_h, (enable == Py_True) ? 1 : 0);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_TCPError);
         return NULL;
@@ -386,6 +345,7 @@ TCP_func_simultaneous_accepts(TCP *self, PyObject *args)
 static PyObject *
 TCP_func_open(TCP *self, PyObject *args)
 {
+    int r;
     long fd;
 
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
@@ -395,7 +355,11 @@ TCP_func_open(TCP *self, PyObject *args)
         return NULL;
     }
 
-    uv_tcp_open((uv_tcp_t *)UV_HANDLE(self), (uv_os_sock_t)fd);
+    r = uv_tcp_open(&self->tcp_h, (uv_os_sock_t)fd);
+    if (r != 0) {
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_TCPError);
+        return NULL;
+    }
 
     Py_RETURN_NONE;
 }
@@ -415,7 +379,7 @@ TCP_tp_init(TCP *self, PyObject *args, PyObject *kwargs)
         return -1;
     }
 
-    r = uv_tcp_init(loop->uv_loop, (uv_tcp_t *)UV_HANDLE(self));
+    r = uv_tcp_init(loop->uv_loop, &self->tcp_h);
     if (r != 0) {
         RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_TCPError);
         return -1;
@@ -431,22 +395,14 @@ static PyObject *
 TCP_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
     TCP *self;
-    uv_tcp_t *uv_tcp;
-
-    uv_tcp = PyMem_Malloc(sizeof *uv_tcp);
-    if (!uv_tcp) {
-        PyErr_NoMemory();
-        return NULL;
-    }
 
     self = (TCP *)StreamType.tp_new(type, args, kwargs);
     if (!self) {
-        PyMem_Free(uv_tcp);
         return NULL;
     }
 
-    uv_tcp->data = (void *)self;
-    UV_HANDLE(self) = (uv_handle_t *)uv_tcp;
+    self->tcp_h.data = self;
+    UV_HANDLE(self) = (uv_handle_t *)&self->tcp_h;
 
     return (PyObject *)self;
 }
