@@ -1,9 +1,11 @@
 
 typedef struct {
+    uv_udp_send_t req;
     PyObject *callback;
     Py_buffer *views;
+    Py_buffer view[1];
     int view_count;
-} udp_send_data_t;
+} udp_send_ctx;
 
 
 static uv_buf_t
@@ -19,9 +21,7 @@ static void
 on_udp_read(uv_udp_t* handle, int nread, uv_buf_t buf, struct sockaddr* addr, unsigned flags)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
-    char ip[INET6_ADDRSTRLEN];
-    struct sockaddr_in addr4;
-    struct sockaddr_in6 addr6;
+    int addrlen;
     uv_err_t err;
     UDP *self;
     PyObject *result, *address_tuple, *data, *py_errorno;
@@ -29,8 +29,7 @@ on_udp_read(uv_udp_t* handle, int nread, uv_buf_t buf, struct sockaddr* addr, un
     ASSERT(handle);
     ASSERT(flags == 0);
 
-    self = (UDP *)handle->data;
-    ASSERT(self);
+    self = PYUV_CONTAINER_OF(handle, UDP, udp_h);
 
     /* Object could go out of scope in the callback, increase refcount to avoid it */
     Py_INCREF(self);
@@ -41,15 +40,8 @@ on_udp_read(uv_udp_t* handle, int nread, uv_buf_t buf, struct sockaddr* addr, un
 
     if (nread > 0) {
         ASSERT(addr);
-        if (addr->sa_family == AF_INET) {
-            addr4 = *(struct sockaddr_in*)addr;
-            uv_ip4_name(&addr4, ip, INET_ADDRSTRLEN);
-            address_tuple = Py_BuildValue("(si)", ip, ntohs(addr4.sin_port));
-        } else {
-            addr6 = *(struct sockaddr_in6*)addr;
-            uv_ip6_name(&addr6, ip, INET6_ADDRSTRLEN);
-            address_tuple = Py_BuildValue("(si)", ip, ntohs(addr6.sin6_port));
-        }
+        addrlen = sizeof(*addr);
+        address_tuple = makesockaddr(addr, addrlen);
         data = PyBytes_FromStringAndSize(buf.base, nread);
         py_errorno = Py_None;
         Py_INCREF(Py_None);
@@ -82,16 +74,15 @@ on_udp_send(uv_udp_send_t* req, int status)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
     int i;
-    udp_send_data_t* req_data;
+    udp_send_ctx *ctx;
     UDP *self;
     PyObject *callback, *result, *py_errorno;
 
     ASSERT(req);
 
-    req_data = (udp_send_data_t *)req->data;
-
-    self = (UDP *)req->handle->data;
-    callback = req_data->callback;
+    ctx = PYUV_CONTAINER_OF(req, udp_send_ctx, req);
+    self = PYUV_CONTAINER_OF(req->handle, UDP, udp_h);
+    callback = ctx->callback;
 
     ASSERT(self);
 
@@ -112,12 +103,13 @@ on_udp_send(uv_udp_send_t* req, int status)
     }
 
     Py_DECREF(callback);
-    for (i = 0; i < req_data->view_count; i++) {
-        PyBuffer_Release(&req_data->views[i]);
+    for (i = 0; i < ctx->view_count; i++) {
+        PyBuffer_Release(&ctx->views[i]);
     }
-    PyMem_Free(req_data->views);
-    PyMem_Free(req_data);
-    PyMem_Free(req);
+    if (ctx->views != ctx->view) {
+        PyMem_Free(ctx->views);
+    }
+    PyMem_Free(ctx);
 
     /* Refcount was increased in the caller function */
     Py_DECREF(self);
@@ -129,32 +121,28 @@ on_udp_send(uv_udp_send_t* req, int status)
 static PyObject *
 UDP_func_bind(UDP *self, PyObject *args)
 {
-    int r, bind_port, address_type, flags;
-    char *bind_ip;
+    int r, flags;
+    struct sockaddr sa;
+    PyObject *addr;
 
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     flags = 0;
 
-    if (!PyArg_ParseTuple(args, "(si)|i:bind", &bind_ip, &bind_port, &flags)) {
+    if (!PyArg_ParseTuple(args, "O|i:bind", &addr, &flags)) {
         return NULL;
     }
 
-    if (bind_port < 0 || bind_port > 65535) {
-        PyErr_SetString(PyExc_ValueError, "port must be between 0 and 65535");
+    if (pyuv_parse_addr_tuple(addr, &sa) < 0) {
+        /* Error is set by the function itself */
         return NULL;
     }
 
-    if (pyuv_guess_ip_family(bind_ip, &address_type)) {
-        PyErr_SetString(PyExc_ValueError, "invalid IP address");
-        return NULL;
-    }
-
-    if (address_type == AF_INET) {
-        r = uv_udp_bind((uv_udp_t *)UV_HANDLE(self), uv_ip4_addr(bind_ip, bind_port), flags);
+    if (sa.sa_family == AF_INET) {
+        r = uv_udp_bind(&self->udp_h, *(struct sockaddr_in *)&sa, flags);
     } else {
-        r = uv_udp_bind6((uv_udp_t *)UV_HANDLE(self), uv_ip6_addr(bind_ip, bind_port), flags);
+        r = uv_udp_bind6(&self->udp_h, *(struct sockaddr_in6 *)&sa, flags);
     }
 
     if (r != 0) {
@@ -186,7 +174,7 @@ UDP_func_start_recv(UDP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_udp_recv_start((uv_udp_t *)UV_HANDLE(self), (uv_alloc_cb)on_udp_alloc, (uv_udp_recv_cb)on_udp_read);
+    r = uv_udp_recv_start(&self->udp_h, (uv_alloc_cb)on_udp_alloc, (uv_udp_recv_cb)on_udp_read);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
@@ -209,7 +197,7 @@ UDP_func_stop_recv(UDP *self)
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    r = uv_udp_recv_stop((uv_udp_t *)UV_HANDLE(self));
+    r = uv_udp_recv_stop(&self->udp_h);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
@@ -225,75 +213,57 @@ UDP_func_stop_recv(UDP *self)
 static PyObject *
 UDP_func_send(UDP *self, PyObject *args)
 {
-    int r, dest_port, address_type;
-    char *dest_ip;
+    int r;
+    struct sockaddr sa;
     uv_buf_t buf;
     Py_buffer *view;
-    PyObject *callback = Py_None;
-    uv_udp_send_t *wr = NULL;
-    udp_send_data_t *req_data = NULL;
+    PyObject *addr, *callback;
+    udp_send_ctx *ctx;
 
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    view = PyMem_Malloc(sizeof *view);
-    if (!view) {
+    callback = Py_None;
+
+    ctx = PyMem_Malloc(sizeof *ctx);
+    if (!ctx) {
         PyErr_NoMemory();
         return NULL;
     }
 
-#ifdef PYUV_PYTHON3
-    if (!PyArg_ParseTuple(args, "(si)y*|O:send", &dest_ip, &dest_port, view, &callback)) {
-#else
-    if (!PyArg_ParseTuple(args, "(si)s*|O:send", &dest_ip, &dest_port, view, &callback)) {
-#endif
+    view = &ctx->view[0];
+
+    if (!PyArg_ParseTuple(args, "O"PYUV_BYTES"*|O:send", &addr, view, &callback)) {
+        PyMem_Free(ctx);
         return NULL;
     }
 
     if (callback != Py_None && !PyCallable_Check(callback)) {
         PyErr_SetString(PyExc_TypeError, "a callable or None is required");
-        goto error1;
+        goto error;
     }
 
-    if (dest_port < 0 || dest_port > 65535) {
-        PyErr_SetString(PyExc_ValueError, "port must be between 0 and 65535");
-        goto error1;
-    }
-
-    if (pyuv_guess_ip_family(dest_ip, &address_type)) {
-        PyErr_SetString(PyExc_ValueError, "invalid IP address");
-        goto error1;
+    if (pyuv_parse_addr_tuple(addr, &sa) < 0) {
+        /* Error is set by the function itself */
+        goto error;
     }
 
     Py_INCREF(callback);
 
-    wr = PyMem_Malloc(sizeof *wr);
-    if (!wr) {
-        PyErr_NoMemory();
-        goto error2;
-    }
-
-    req_data = PyMem_Malloc(sizeof *req_data);
-    if (!req_data) {
-        PyErr_NoMemory();
-        goto error2;
-    }
-
     buf = uv_buf_init(view->buf, view->len);
-    req_data->callback = callback;
-    req_data->view_count = 1;
-    req_data->views = view;
+    ctx->callback = callback;
+    ctx->view_count = 1;
+    ctx->views = view;
 
-    wr->data = (void *)req_data;
-
-    if (address_type == AF_INET) {
-        r = uv_udp_send(wr, (uv_udp_t *)UV_HANDLE(self), &buf, 1, uv_ip4_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
+    if (sa.sa_family == AF_INET) {
+        r = uv_udp_send(&ctx->req, &self->udp_h, &buf, 1, *(struct sockaddr_in *)&sa, (uv_udp_send_cb)on_udp_send);
     } else {
-        r = uv_udp_send6(wr, (uv_udp_t *)UV_HANDLE(self), &buf, 1, uv_ip6_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
+        r = uv_udp_send6(&ctx->req, &self->udp_h, &buf, 1, *(struct sockaddr_in6 *)&sa, (uv_udp_send_cb)on_udp_send);
     }
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
-        goto error2;
+        Py_DECREF(callback);
+        goto error;
     }
 
     /* Increase refcount so that object is not removed before the callback is called */
@@ -301,13 +271,9 @@ UDP_func_send(UDP *self, PyObject *args)
 
     Py_RETURN_NONE;
 
-error2:
-    Py_DECREF(callback);
-    PyMem_Free(req_data);
-    PyMem_Free(wr);
-error1:
+error:
     PyBuffer_Release(view);
-    PyMem_Free(view);
+    PyMem_Free(ctx);
     return NULL;
 }
 
@@ -315,20 +281,20 @@ error1:
 static PyObject *
 UDP_func_sendlines(UDP *self, PyObject *args)
 {
-    int i, r, buf_count, dest_port, address_type;
-    char *dest_ip;
-    PyObject *callback, *seq;
+    int i, r, buf_count;
+    struct sockaddr sa;
+    PyObject *addr, *callback, *seq;
     Py_buffer *views;
     uv_buf_t *bufs;
-    uv_udp_send_t *wr = NULL;
-    udp_send_data_t *req_data = NULL;
+    udp_send_ctx *ctx;
 
     callback = Py_None;
+    ctx = NULL;
 
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    if (!PyArg_ParseTuple(args, "(si)O|O:sendlines", &dest_ip, &dest_port, &seq, &callback)) {
+    if (!PyArg_ParseTuple(args, "OO|O:sendlines", &addr, &seq, &callback)) {
         return NULL;
     }
 
@@ -337,13 +303,8 @@ UDP_func_sendlines(UDP *self, PyObject *args)
         return NULL;
     }
 
-    if (dest_port < 0 || dest_port > 65535) {
-        PyErr_SetString(PyExc_ValueError, "port must be between 0 and 65535");
-        return NULL;
-    }
-
-    if (pyuv_guess_ip_family(dest_ip, &address_type)) {
-        PyErr_SetString(PyExc_ValueError, "invalid IP address");
+    if (pyuv_parse_addr_tuple(addr, &sa) < 0) {
+        /* Error is set by the function itself */
         return NULL;
     }
 
@@ -355,27 +316,20 @@ UDP_func_sendlines(UDP *self, PyObject *args)
 
     Py_INCREF(callback);
 
-    wr = PyMem_Malloc(sizeof *wr);
-    if (!wr) {
+    ctx = PyMem_Malloc(sizeof *ctx);
+    if (!ctx) {
         PyErr_NoMemory();
         goto error;
     }
 
-    req_data = PyMem_Malloc(sizeof *req_data);
-    if (!req_data) {
-        PyErr_NoMemory();
-        goto error;
-    }
+    ctx->callback = callback;
+    ctx->view_count = buf_count;
+    ctx->views = views;
 
-    req_data->callback = callback;
-    req_data->view_count = buf_count;
-    req_data->views = views;
-    wr->data = (void *)req_data;
-
-    if (address_type == AF_INET) {
-        r = uv_udp_send(wr, (uv_udp_t *)UV_HANDLE(self), bufs, buf_count, uv_ip4_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
+    if (sa.sa_family == AF_INET) {
+        r = uv_udp_send(&ctx->req, &self->udp_h, bufs, buf_count, *(struct sockaddr_in *)&sa, (uv_udp_send_cb)on_udp_send);
     } else {
-        r = uv_udp_send6(wr, (uv_udp_t *)UV_HANDLE(self), bufs, buf_count, uv_ip6_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
+        r = uv_udp_send6(&ctx->req, &self->udp_h, bufs, buf_count, *(struct sockaddr_in6 *)&sa, (uv_udp_send_cb)on_udp_send);
     }
 
     /* uv_write copies the uv_buf_t structures, so we can free them now */
@@ -399,8 +353,7 @@ error:
     }
     PyMem_Free(views);
     PyMem_Free(bufs);
-    PyMem_Free(req_data);
-    PyMem_Free(wr);
+    PyMem_Free(ctx);
     return NULL;
 }
 
@@ -420,7 +373,7 @@ UDP_func_set_membership(UDP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_udp_set_membership((uv_udp_t *)UV_HANDLE(self), multicast_address, interface_address, membership);
+    r = uv_udp_set_membership(&self->udp_h, multicast_address, interface_address, membership);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
@@ -434,34 +387,20 @@ static PyObject *
 UDP_func_getsockname(UDP *self)
 {
     int r, namelen;
-    char ip[INET6_ADDRSTRLEN];
     struct sockaddr sockname;
-    struct sockaddr_in *addr4;
-    struct sockaddr_in6 *addr6;
 
     namelen = sizeof(sockname);
 
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    r = uv_udp_getsockname((uv_udp_t *)UV_HANDLE(self), &sockname, &namelen);
+    r = uv_udp_getsockname(&self->udp_h, &sockname, &namelen);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
     }
 
-    if (sockname.sa_family == AF_INET) {
-        addr4 = (struct sockaddr_in*)&sockname;
-        uv_ip4_name(addr4, ip, INET_ADDRSTRLEN);
-        return Py_BuildValue("si", ip, ntohs(addr4->sin_port));
-    } else if (sockname.sa_family == AF_INET6) {
-        addr6 = (struct sockaddr_in6*)&sockname;
-        uv_ip6_name(addr6, ip, INET6_ADDRSTRLEN);
-        return Py_BuildValue("si", ip, ntohs(addr6->sin6_port));
-    } else {
-        PyErr_SetString(PyExc_UDPError, "unknown address type detected");
-        return NULL;
-    }
+    return makesockaddr(&sockname, namelen);
 }
 
 
@@ -482,7 +421,7 @@ UDP_func_set_multicast_ttl(UDP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_udp_set_multicast_ttl((uv_udp_t *)UV_HANDLE(self), ttl);
+    r = uv_udp_set_multicast_ttl(&self->udp_h, ttl);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
@@ -505,7 +444,7 @@ UDP_func_set_broadcast(UDP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_udp_set_broadcast((uv_udp_t *)UV_HANDLE(self), (enable == Py_True) ? 1 : 0);
+    r = uv_udp_set_broadcast(&self->udp_h, (enable == Py_True) ? 1 : 0);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
@@ -528,7 +467,7 @@ UDP_func_set_multicast_loop(UDP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_udp_set_multicast_loop((uv_udp_t *)UV_HANDLE(self), (enable == Py_True) ? 1 : 0);
+    r = uv_udp_set_multicast_loop(&self->udp_h, (enable == Py_True) ? 1 : 0);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
@@ -555,7 +494,7 @@ UDP_func_set_ttl(UDP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_udp_set_ttl((uv_udp_t *)UV_HANDLE(self), ttl);
+    r = uv_udp_set_ttl(&self->udp_h, ttl);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
@@ -577,7 +516,7 @@ UDP_func_open(UDP *self, PyObject *args)
         return NULL;
     }
 
-    uv_udp_open((uv_udp_t *)UV_HANDLE(self), (uv_os_sock_t)fd);
+    uv_udp_open(&self->udp_h, (uv_os_sock_t)fd);
 
     Py_RETURN_NONE;
 }
@@ -597,7 +536,7 @@ UDP_tp_init(UDP *self, PyObject *args, PyObject *kwargs)
         return -1;
     }
 
-    r = uv_udp_init(loop->uv_loop, (uv_udp_t *)UV_HANDLE(self));
+    r = uv_udp_init(loop->uv_loop, &self->udp_h);
     if (r != 0) {
         RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_UDPError);
         return -1;
@@ -613,22 +552,14 @@ static PyObject *
 UDP_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
     UDP *self;
-    uv_udp_t *uv_udp;
-
-    uv_udp = PyMem_Malloc(sizeof *uv_udp);
-    if (!uv_udp) {
-        PyErr_NoMemory();
-        return NULL;
-    }
 
     self = (UDP *)HandleType.tp_new(type, args, kwargs);
     if (!self) {
-        PyMem_Free(uv_udp);
         return NULL;
     }
 
-    uv_udp->data = (void *)self;
-    UV_HANDLE(self) = (uv_handle_t *)uv_udp;
+    self->udp_h.data = self;
+    UV_HANDLE(self) = (uv_handle_t *)&self->udp_h;
 
     return (PyObject *)self;
 }
