@@ -84,8 +84,8 @@ Util_func_resident_set_memory(PyObject *obj)
 static PyObject *
 Util_func_interface_addresses(PyObject *obj)
 {
+    static char buf[INET6_ADDRSTRLEN+1];
     int i, count;
-    char ip[INET6_ADDRSTRLEN];
     uv_interface_address_t* interfaces;
     uv_err_t err;
     PyObject *result, *item, *exc_data;
@@ -109,11 +109,11 @@ Util_func_interface_addresses(PyObject *obj)
             PyStructSequence_SET_ITEM(item, 0, Py_BuildValue("s", interfaces[i].name));
             PyStructSequence_SET_ITEM(item, 1, PyBool_FromLong((long)interfaces[i].is_internal));
             if (interfaces[i].address.address4.sin_family == AF_INET) {
-                uv_ip4_name(&interfaces[i].address.address4, ip, INET_ADDRSTRLEN);
+                uv_ip4_name(&interfaces[i].address.address4, buf, sizeof(buf));
             } else if (interfaces[i].address.address4.sin_family == AF_INET6) {
-                uv_ip6_name(&interfaces[i].address.address6, ip, INET6_ADDRSTRLEN);
+                uv_ip6_name(&interfaces[i].address.address6, buf, sizeof(buf));
             }
-            PyStructSequence_SET_ITEM(item, 2, Py_BuildValue("s", ip));
+            PyStructSequence_SET_ITEM(item, 2, Py_BuildValue("s", buf));
             PyList_SET_ITEM(result, i, item);
         }
         uv_free_interface_addresses(interfaces, count);
@@ -179,56 +179,22 @@ Util_func_cpu_info(PyObject *obj)
 }
 
 
-/* Modified from Python Modules/socketmodule.c */
-static PyObject *
-makesockaddr(struct sockaddr *addr, int addrlen)
-{
-    struct sockaddr_in *addr4;
-    struct sockaddr_in6 *addr6;
-    char ip[INET6_ADDRSTRLEN];
-
-    if (addrlen == 0) {
-        /* No address */
-        Py_RETURN_NONE;
-    }
-
-    switch (addr->sa_family) {
-    case AF_INET:
-    {
-        addr4 = (struct sockaddr_in*)addr;
-        uv_ip4_name(addr4, ip, INET_ADDRSTRLEN);
-        return Py_BuildValue("si", ip, ntohs(addr4->sin_port));
-    }
-
-    case AF_INET6:
-    {
-        addr6 = (struct sockaddr_in6*)addr;
-        uv_ip6_name(addr6, ip, INET6_ADDRSTRLEN);
-        return Py_BuildValue("siii", ip, ntohs(addr6->sin6_port), addr6->sin6_flowinfo, addr6->sin6_scope_id);
-    }
-
-    default:
-        /* If we don't know the address family, don't raise an exception -- return it as a tuple. */
-        return Py_BuildValue("is#", addr->sa_family, addr->sa_data, sizeof(addr->sa_data));
-    }
-}
-
 static void
 getaddrinfo_cb(uv_getaddrinfo_t* req, int status, struct addrinfo* res)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
     struct addrinfo *ptr;
-    uv_err_t err;
     Loop *loop;
-    GAIRequest *pyreq;
+    GAIRequest *gai_req;
     PyObject *addr, *item, *errorno, *dns_result, *result;
 
     ASSERT(req);
-    pyreq = (GAIRequest *)req->data;
-    loop = (Loop *)req->loop->data;
+
+    gai_req = PYUV_CONTAINER_OF(req, GAIRequest, req);
+    loop = REQUEST(gai_req)->loop;
 
     if (status != 0) {
-        err = uv_last_error(loop->uv_loop);
+        uv_err_t err = uv_last_error(req->loop);
         errorno = PyInt_FromLong((long)err.code);
         dns_result = Py_None;
         Py_INCREF(Py_None);
@@ -268,7 +234,7 @@ getaddrinfo_cb(uv_getaddrinfo_t* req, int status, struct addrinfo* res)
     Py_INCREF(Py_None);
 
 callback:
-    result = PyObject_CallFunctionObjArgs(pyreq->callback, dns_result, errorno, NULL);
+    result = PyObject_CallFunctionObjArgs(gai_req->callback, dns_result, errorno, NULL);
     if (result == NULL) {
         handle_uncaught_exception(loop);
     }
@@ -277,13 +243,10 @@ callback:
     Py_DECREF(errorno);
 
     Py_DECREF(loop);
-    Py_DECREF(pyreq->callback);
-    pyreq->callback = NULL;
-    ((Request *)pyreq)->req = NULL;
-    Py_DECREF(pyreq);
+    Py_DECREF(gai_req->callback);
+    Py_DECREF(gai_req);
 
     uv_freeaddrinfo(res);
-    PyMem_Free(req);
 
     PyGILState_Release(gstate);
 }
@@ -295,16 +258,14 @@ Util_func_getaddrinfo(PyObject *obj, PyObject *args, PyObject *kwargs)
     char port_str[6];
     int port, family, socktype, protocol, flags, r;
     struct addrinfo hints;
-    uv_getaddrinfo_t* req;
     Loop *loop;
-    GAIRequest *pyreq;
+    GAIRequest *gai_req;
     PyObject *callback, *host, *idna;
 
     static char *kwlist[] = {"loop", "callback", "host", "port", "family", "socktype", "protocol", "flags", NULL};
 
     UNUSED_ARG(obj);
-    req = NULL;
-    pyreq = NULL;
+    gai_req = NULL;
     port = socktype = protocol = flags = 0;
     family = AF_UNSPEC;
 
@@ -337,23 +298,16 @@ Util_func_getaddrinfo(PyObject *obj, PyObject *args, PyObject *kwargs)
     }
     snprintf(port_str, sizeof(port_str), "%d", port);
 
-    req = PyMem_Malloc(sizeof *req);
-    if (!req) {
+    gai_req = (GAIRequest *)PyObject_CallObject((PyObject *)&GAIRequestType, NULL);
+    if (!gai_req) {
         PyErr_NoMemory();
-        goto error;
+        return NULL;
     }
 
-    pyreq = (GAIRequest *)PyObject_CallObject((PyObject *)&GAIRequestType, NULL);
-    if (!pyreq) {
-        PyErr_NoMemory();
-        goto error;
-    }
-    ((Request *)pyreq)->req = (uv_req_t *)req;
-    pyreq->callback = callback;
-
+    gai_req->callback = callback;
+    REQUEST(gai_req)->loop = loop;
     Py_INCREF(loop);
     Py_INCREF(callback);
-    req->data = (void *)pyreq;
 
     memset(&hints, 0, sizeof(hints));
     hints.ai_family = family;
@@ -361,18 +315,19 @@ Util_func_getaddrinfo(PyObject *obj, PyObject *args, PyObject *kwargs)
     hints.ai_protocol = protocol;
     hints.ai_flags = flags;
 
-    r = uv_getaddrinfo(loop->uv_loop, req, &getaddrinfo_cb, host_str, port_str, &hints);
+    r = uv_getaddrinfo(loop->uv_loop, &gai_req->req, &getaddrinfo_cb, host_str, port_str, &hints);
     if (r != 0) {
         RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_UVError);
         goto error;
     }
 
-    Py_INCREF(pyreq);
-    return (PyObject *)pyreq;
+    Py_INCREF(gai_req);
+    return (PyObject *)gai_req;
 
 error:
-    Py_XDECREF(pyreq);
-    PyMem_Free(req);
+    Py_DECREF(gai_req);
+    Py_DECREF(loop);
+    Py_DECREF(callback);
     return NULL;
 }
 
@@ -428,9 +383,8 @@ check_signals(uv_poll_t *handle, int status, int events)
     SignalChecker *self;
 
     ASSERT(handle);
-    self = (SignalChecker *)handle->data;
 
-    Py_INCREF(self);
+    self = PYUV_CONTAINER_OF(handle, SignalChecker, poll_h);
 
     if (status == 0) {
         ASSERT(events == UV_READABLE);
@@ -461,7 +415,7 @@ SignalChecker_func_start(SignalChecker *self)
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    r = uv_poll_start((uv_poll_t *)UV_HANDLE(self), UV_READABLE, check_signals);
+    r = uv_poll_start(&self->poll_h, UV_READABLE, check_signals);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UVError);
         return NULL;
@@ -479,7 +433,7 @@ SignalChecker_func_stop(SignalChecker *self)
     RAISE_IF_HANDLE_NOT_INITIALIZED(self, NULL);
     RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    r = uv_poll_stop((uv_poll_t *)UV_HANDLE(self));
+    r = uv_poll_stop(&self->poll_h);
     if (r != 0) {
         RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UVError);
         return NULL;
@@ -504,7 +458,7 @@ SignalChecker_tp_init(SignalChecker *self, PyObject *args, PyObject *kwargs)
         return -1;
     }
 
-    r = uv_poll_init_socket(loop->uv_loop, (uv_poll_t *)UV_HANDLE(self), (uv_os_sock_t)fd);
+    r = uv_poll_init_socket(loop->uv_loop, &self->poll_h, (uv_os_sock_t)fd);
     if (r != 0) {
         RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_UVError);
         return -1;
@@ -522,22 +476,14 @@ static PyObject *
 SignalChecker_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
     SignalChecker *self;
-    uv_poll_t *uv_poll;
-
-    uv_poll = PyMem_Malloc(sizeof *uv_poll);
-    if (!uv_poll) {
-        PyErr_NoMemory();
-        return NULL;
-    }
 
     self = (SignalChecker *)HandleType.tp_new(type, args, kwargs);
     if (!self) {
-        PyMem_Free(uv_poll);
         return NULL;
     }
 
-    uv_poll->data = (void *)self;
-    UV_HANDLE(self) = (uv_handle_t *)uv_poll;
+    self->poll_h.data = self;
+    UV_HANDLE(self) = (uv_handle_t *)&self->poll_h;
 
     return (PyObject *)self;
 }
@@ -648,4 +594,3 @@ init_util(void)
     return module;
 }
 
-
