@@ -13,8 +13,13 @@ def on_write2(handle, error):
 
 def on_channel_read(handle, data, pending, error):
     global channel, loop, recv_handle
-    assert pending == pyuv.UV_NAMED_PIPE, "wrong handle"
-    recv_handle = pyuv.Pipe(loop)
+    assert pending in (pyuv.UV_NAMED_PIPE, pyuv.UV_UDP, pyuv.UV_TCP), "wrong handle type"
+    if pending == pyuv.UV_NAMED_PIPE:
+        recv_handle = pyuv.Pipe(loop)
+    elif pending == pyuv.UV_TCP:
+        recv_handle = pyuv.TCP(loop)
+    elif pending == pyuv.UV_UDP:
+        recv_handle = pyuv.UDP(loop)
     channel.accept(recv_handle)
     channel.write2(b".", recv_handle, on_write2)
 
