@@ -1,6 +1,8 @@
 
-from common import platform_skip, unittest2
+from common import unittest2
+
 import pyuv
+import socket
 
 
 class UtilTest(unittest2.TestCase):
@@ -37,12 +39,12 @@ class UtilTest(unittest2.TestCase):
         r = pyuv.util.cpu_info()
         self.assertTrue(r)
 
-    @platform_skip(['darwin'])
-    def test_process_title(self):
-        title = 'my process'
-        pyuv.util.set_process_title(title)
-        r = pyuv.util.get_process_title()
-        self.assertEqual(r, title)
+    def test_getaddrinfo(self):
+        def getaddrinfo_cb(result, errorno):
+            self.assertEqual(errorno, None)
+        loop = pyuv.Loop.default_loop()
+        pyuv.util.getaddrinfo(loop, 'localhost', getaddrinfo_cb, 80, socket.AF_INET)
+        loop.run()
 
 
 if __name__ == '__main__':
